<?php
namespace dsmirror;
require_once 'ReplicationServiceClient.php';

/**
 *
 * @author Tecknolab
 *        
 */
class ReplicationManager
{

    /**
     */
    public function __construct($replicationService)
    {
        $this->repService = $replicationService;
    }
    
    public function __destruct()
    {
        unset($this->repService);
    }
    
    public function serverProps() /*: Array of properties */
    {
        return $this->repService->properties();
    }
    
    public function listAvailableDSConfig() /*: iterable Array of DSConfig*/
    {
        return $this->repService->getDSList();
    }
    
    public function registerReverseDSConfig($db, $conn) /*: iterable Array of DSConfig*/
    {
        $sql = $this->repService->generateCandidateReverseDSConfigScript(NULL);
        $stmt = $conn->prepare($sql);
        $stmt->bind_param("s", $db);
        $stmt->execute();
        if (!($rset = $stmt->get_result())) {
            throw new \Exception("Error executing hash query:".$sql." Error:".$conn->error);
        }
        $metadata = Array();
        $table = NULL;
        $tableMetadata = Array();
        if ($rset->num_rows > 0) {
            while($row = $rset->fetch_assoc()) {
                if (isset($row["TABLE_NAME"]) && $table==NULL){
                    $table = $row["TABLE_NAME"];
                } else if (isset($row["TABLE_NAME"]) && $table!=$row["TABLE_NAME"]) {                    
                    $metadata[$table]=$tableMetadata;
                    $table=$row["TABLE_NAME"];
                    $tableMetadata=Array();
                }
                $tableMetadata[]=$row;
            }
            if ($table != NULL)
                $metadata[$table]=$tableMetadata;
        }
        $rset->close();
        $stmt->close();
        return $this->repService->registerCandidateReverseDSConfig($metadata);
    }
    
    public function listAvailableReverseDSConfig() /*: iterable Array of DSConfig*/
    {
        return $this->repService->getDSListReverse();
    }
    
    public function readDSConfig($dsConfigId) /*: DSConfig */
    {
        foreach ($this->listAvailableDSConfig() as $dsConfig)
        {
            if ($dsConfig->id == $dsConfigId) {
                return $dsConfig;
            }
        }
        foreach ($this->listAvailableReverseDSConfig() as $dsConfig)
        {
            if ($dsConfig->id == $dsConfigId) {
                return $dsConfig;
            }
        }
        throw new \Exception("Datasource config id ".$dsConfigId." not found in server");
    }
    
    public function createReplica($dsConfigId, $dsReplicaOptions) /*: DSReplica */
    {
        return $this->repService->registerDSReplica($dsConfigId, $dsReplicaOptions);
    }
    
    public function runReplica($conn, $replicaExec) /*: ReplicaExec */
    {
        $res = new ReplicaExec($replicaExec);
        trigger_error("Requested execution of ReplicaExec ". json_encode($res), E_USER_NOTICE);
        $res->timedout = false;
        $dsReplica = $res->dsReplica;
        if (!isset($res->dsConfig) || empty($res->dsConfig->id)) {
            throw new \Exception("Invalid or null datasource config id");            
        }
        $initRun = time();
        $timeoutSeconds = $replicaExec->timeout;
        if (isset($dsReplica) && isset($dsReplica->dsConfigId) && $dsReplica->dsConfigId !== $res->dsConfig->id) {
            throw new \Exception("Inconsistent replica, declared for dsConfigId ". $dsReplica->dsConfigId . " and configured in execution for ". $res->dsConfig->id);
        }
        $dsConfig = $res->dsConfig;
        if (!isset($dsReplica) || empty($dsReplica->dsReplicaId)) {
            // create replica
            $res->dsReplica = $this->repService->registerDSReplica($dsConfig->id, $dsReplica->options);
        }
        if (isset($res->jobId) && $replicaExec->timedout) {
            try {
                trigger_error("Replica ". $res->dsReplicaId." timedout in previous execution. Trying to retrieve job ".$res->jobId , E_USER_NOTICE);
                $context = $this->repService->getReplicationContext($res->jobId);
                trigger_error("Retrieved context ".json_encode($context) , E_USER_NOTICE);
                $replicaJob = $context->pending;
                $segmentHashDescriptorArray = $replicaJob->segmentsList;
                $lastPartialSegment = $context->lastPartialSegment;
                $nextPartialQuery = !empty($context->nextPartialFilter) ? $replicaJob->selectQuery . $context->nextPartialFilter . $replicaJob->orderBy : NULL;
            } catch (RESTException $e)
            {
                if ($e->getCode() == 404) {
                    // replica job not found in server, continue the regular path
                    trigger_error("Job ". $res->jobId." not found in server. Starting a new one ", E_USER_NOTICE);
                    $res->jobId = NULL;
                } else {
                    throw $e;
                }
            }
        }
        if ($res->jobId == NULL) {
            try {
                $replicaJob = $this->repService->startSyncDS($res->dsReplica->dsReplicaId);
                $res->jobId = $replicaJob->jobId;
            } catch (RESTException $e)
            {
                if ($e->getCode() == 404) {
                    // replica not found in server, create one
                    trigger_error("Replica ". $res->dsReplicaId." not found in server. Starting a new one ", E_USER_NOTICE);
                    $res->dsReplica = $this->repService->registerDSReplica($dsConfig->id, $dsReplica->options);
                    $replicaJob = $this->repService->startSyncDS($res->dsReplica->dsReplicaId);
                    $res->jobId = $replicaJob->jobId;
                } else {
                    throw $e;
                }
            }
        }
        trigger_error("Replica job received". json_encode($replicaJob), E_USER_NOTICE);
        if ($this->timedout($initRun, $timeoutSeconds)) {$res->timedout=true;return $res;}
        try {
            if (!$dsConfig->reverse && !$this->validateTargetTable($conn, $dsConfig, !empty($res->dsReplica->options->targetTableName) ? $res->dsReplica->options->targetTableName : $dsConfig->tableName)) {
                if ($res->dsReplica->options->createTable)
                {
                    $this->createTarget($conn, $res->dsReplica);
                }
                else
                {
                    throw new \Exception("Target table ". $res->dsReplica->options->targetTableName . " does not exist in local database");
                }
            }
            if (!isset($segmentHashDescriptorArray)) {
                $segmentHashDescriptorArray = $this->calcDirtySegments($conn, $replicaJob);
            } else {
                trigger_error("Continuing with pending replication segments of #". count($segmentHashDescriptorArray), E_USER_NOTICE);                
            }
            $success = true;
            $replicationError = NULL;
            if ($this->timedout($initRun, $timeoutSeconds)) {$res->timedout=true;return $res;}
            for ($j = 0; $j < count($segmentHashDescriptorArray); ++$j)
            {
                $res->timedout = $this->getAndApplyDelta($conn, $segmentHashDescriptorArray[$j], $replicaJob, $dsConfig, $lastPartialSegment, $nextPartialQuery, $initRun, $timeoutSeconds);
                if ($res->timedout) {return $res;}
            }
            if (!$res->timedout) {
                $this->repService->finishSyncDS($replicaJob->jobId);
                unset($res->jobId);
                if (isset($pendingPartialSegment)) unset($pendingPartialSegment);
            }
            // TODO. Handle potential segment errors
        } catch (\Exception $e) {
            if (isset($res->jobId)) {try {$this->repService->finishSyncDS($replicaJob->jobId);} catch (\Exception $ex) {};}
            throw $e;
        }
        return $res;
    }
    
    public function deleteReplica($dsReplica)
    {
        if (!empty($dsReplica->dsReplicaId)) {
            // delete replica
            $res = $this->repService->deleteReplica($dsReplica->dsReplicaId);
        }
    }
    
    private function calcDirtySegments($conn, $replicaJob) /*: iterable*/ // Array of SegmentHashDescriptor
    {
        for ($index = 0; $index < count($replicaJob->segmentsList); ++$index) {
            trigger_error("Calculating segment hash for segment ". json_encode($replicaJob->segmentsList[$index]->segment)." with query ". $replicaJob->segmentHashQuery . $replicaJob->segmentsList[$index]->segmentFilter, E_USER_NOTICE);
            if (!($rset = $conn->query($replicaJob->segmentHashQuery . $replicaJob->segmentsList[$index]->segmentFilter))) {
                throw new \Exception("Error executing hash query:".$replicaJob->segmentsList[$index]->hashQuery." Error:".$conn->error);
            }
            trigger_error("Segment hash query return ".$rset->num_rows." rows", E_USER_NOTICE);
            if ($rset->num_rows > 0) {
                $obj = $rset->fetch_object();
                trigger_error("segment size ".$obj->SIZE." hash: ".$obj->HASH, E_USER_NOTICE);
            } else {
                throw new \Exception("Hash query does not return results. ".$replicaJob->segmentHashQuery . $replicaJob->segmentsList[$index]->segmentFilter);
            }
            $rset->close();
            $segmentHashArray[$index] = new SegmentHash();
            $segmentHashArray[$index]->segment = $replicaJob->segmentsList[$index]->segment;
            $segmentHashArray[$index]->segment->size = $obj->SIZE;
            $segmentHashArray[$index]->hash = $obj->HASH;
        }
        trigger_error("Requesting dirty segments for job ".$replicaJob->jobId, E_USER_NOTICE);
        $segmentHashDescriptorArray = $this->repService->getReplicationDirtySegments($replicaJob->jobId, $segmentHashArray);
        return $segmentHashDescriptorArray;
    }
    
    private function getAndApplyDelta($conn, $segmentHashDescriptor, $replicaJob, $dsConfig, $lastPartialSegment, $nextPartialQuery, $initRun=NULL, $timeoutSeconds=NULL)
    {
        $timeout = $this->timedout($initRun, $timeoutSeconds);
        if ($timeout) return $timeout;
        $replicationDeltaResult = new ReplicationDeltaResult();
        $segmentContentHashResult = new SegmentContentHashResult();
        $segmentContentHashResult->segment = $segmentHashDescriptor->segment;
        $segmentContentHashResult->partialSegment = $lastPartialSegment;
        $segmentQuery = !empty($nextPartialQuery) ? $nextPartialQuery :$replicaJob->selectQuery . $segmentHashDescriptor->segmentFilter . $replicaJob->orderBy;
        trigger_error("Executing query for calculating content hash ".$segmentQuery. " with maxSize ".(isset($replicaJob->maxContentRowsPerRequest) ? $replicaJob->maxContentRowsPerRequest : "undefined"), E_USER_NOTICE);
        $firstCycle=!isset($lastPartialSegment);
        $segmentFullyProcessed = false;
        $moreContentToFetch = true;
        do {
            $timeout = $this->timedout($initRun, $timeoutSeconds);
            if ($timeout) break;
            if ($firstCycle && $dsConfig->reverse) {
                trigger_error("Calculating new segment hash", E_USER_NOTICE);
                if (!($hashRset = $conn->query($replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter))) {
                    throw new \Exception("Error executing hash query:".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter. " Error:".$conn->error);
                }
                trigger_error("Segment hash query return ".$hashRset->num_rows." rows", E_USER_NOTICE);
                if ($hashRset->num_rows > 0) {
                    $obj = $hashRset->fetch_object();
                    trigger_error("new segment size ".$obj->SIZE." hash: ".$obj->HASH, E_USER_NOTICE);
                } else {
                    throw new \Exception("Hash query does not return results. ".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter);
                }
                $hashRset->close();unset($hashRset);
                $segmentContentHashResult->hashClient=$obj->HASH;
            }
            if (!($rset = $conn->query($segmentQuery))) {
                throw new \Exception("Error retrieving table data for hash. query:".$segmentQuery." Error:".$conn->error);
            }
            $segmentContentHashResult->rowHashList = $this->calculateSegmentContentHash($rset, $dsConfig, $replicaJob->maxContentRowsPerRequest);
            $rset->close(); unset($rset);
            trigger_error("Fetched ".count($segmentContentHashResult->rowHashList)." rows", E_USER_NOTICE);
            if (isset($segmentContentHashResult->partialSegment))
            {
                trigger_error("Rolling over partial Segment ".json_encode($segmentContentHashResult->partialSegment), E_USER_NOTICE);
                if (!isset($rollingSegment)) {
                    $segmentContentHashResult->partialSegment->pkIniExcluded = $segmentContentHashResult->partialSegment->pkEndIncluded;
                    if (count($segmentContentHashResult->rowHashList) >= $replicaJob->maxContentRowsPerRequest) {
                        $segmentContentHashResult->partialSegment->pkEndIncluded = end($segmentContentHashResult->rowHashList)->row;
                    } else {
                        $segmentContentHashResult->partialSegment->pkEndIncluded = $segmentContentHashResult->segment->pkEndIncluded;
                    }
                } else {
                    // rolling segment already calculated from server partial segment
                    $segmentContentHashResult->partialSegment = $rollingSegment;
                }
                $segmentContentHashResult->partialSegment->size=count($segmentContentHashResult->rowHashList);
                trigger_error("New partial Segment ".json_encode($segmentContentHashResult->partialSegment), E_USER_NOTICE);
            }
            if (isset($replicaJob->maxContentRowsPerRequest) && count($segmentContentHashResult->rowHashList) >= $replicaJob->maxContentRowsPerRequest)
            {
                if (!isset($segmentContentHashResult->partialSegment)) {
                    $segmentContentHashResult->partialSegment = new Segment(NULL);
                    $segmentContentHashResult->partialSegment->pkIniExcluded = $segmentContentHashResult->segment->pkIniIncluded;
                }
                $segmentContentHashResult->partialSegment->pkEndIncluded = end($segmentContentHashResult->rowHashList)->row;
                $segmentContentHashResult->partialSegment->size=count($segmentContentHashResult->rowHashList);
                trigger_error("Declaring new partial Segment ".json_encode($segmentContentHashResult->partialSegment), E_USER_NOTICE);
            }
            $rollingSegment = NULL;
            $replicationDelta = $this->repService->getSegmentReplicationDelta($replicaJob->jobId, $segmentContentHashResult);
            if (isset($replicationDelta->partialSegment)) {
                $segmentQuery = $replicaJob->selectQuery . $replicationDelta->nextPartialFilter;
                trigger_error("Partial segment delta. next query ".$replicationDelta->nextPartialQuery, E_USER_NOTICE);
                if (!isset($segmentContentHashResult->partialSegment) || $replicationDelta->partialSegment->pkEndIncluded !== $segmentContentHashResult->partialSegment->pkEndIncluded) {
                    trigger_error("Declared new partial Segment from server. ".json_encode($replicationDelta->partialSegment), E_USER_NOTICE);
                    $rollingSegment = new Segment(NULL);
                    $rollingSegment->pkIniExcluded = $replicationDelta->partialSegment->pkEndIncluded;
                    $rollingSegment->pkEndIncluded = isset($segmentContentHashResult->partialSegment) ? $segmentContentHashResult->partialSegment->pkEndIncluded : $segmentContentHashResult->segment->pkEndIncluded;
                    $segmentContentHashResult->partialSegment = $replicationDelta->partialSegment;
                }
            }
            trigger_error("Received replication delta content with ".(isset($replicationDelta->commandList) ? count($replicationDelta->commandList) : "0")." items.".(isset($replicationDelta->partialSegment) ? " Corresponding to partialSegment ".json_encode($replicationDelta->partialSegment):""), E_USER_NOTICE);
            $segmentFullyProcessed = !isset($segmentContentHashResult->partialSegment) ||
                $segmentContentHashResult->partialSegment->pkEndIncluded === $segmentContentHashResult->segment->pkEndIncluded;
            $replicationDeltaResult->operationResult = Array();
            if (isset($replicationDelta->commandList)) {
                foreach ($replicationDelta->commandList as $command) {
                    if ($dsConfig->reverse) {
                        if ($command->operation === "SELECT_I" || $command->operation === "SELECT_U") {
                            if (!($rsetComm = $conn->query($replicaJob->selectQuery . $command->target))) {
                                //TODO. abort replication and send error to server
                                throw new \Exception("Error executing outbound replica delta operation:".$command->operation." ".$replicaJob->selectQuery . $command->target." Error: ".$conn->error);
                            }
                            if ($rsetComm->num_rows > 0) {
                                $replicationDeltaResultItem = new ReplicationDeltaResultItem();
                                $replicationDeltaResultItem->operation = $command->operation;
                                $replicationDeltaResultItem->row = $rsetComm->fetch_row();
                                $replicationDeltaResult->operationResult[] = $replicationDeltaResultItem;
                            }
                            $rsetComm->close();
                        } else {
                            throw new \Exception("Invalid command received for outbound replication: ".$command->operation);
                        }
                    } else {
                        $stmt = "";
                        if ($command->operation == "INSERT") {$stmt = $replicaJob->insertStmt;}
                        if ($command->operation == "UPDATE") {$stmt = $replicaJob->updateStmt;}
                        if ($command->operation == "DELETE") {$stmt = $replicaJob->deleteStmt;}
                        if (empty($stmt)) {
                            throw new \Exception("Invalid command received for inbound replication: ".$command->operation);    
                        }
                        if (!$conn->query($command->operation." ".$stmt.$command->target)) {
                            //TODO. abort replication and send error to server
                            throw new \Exception("Error executing delta operation:".$command->operation." ".$stmt.$command->target." Error: ".$conn->error);
                        }
                    }
                }
            }
            $segmentHash = new SegmentHash();
            $segmentHash->segment = $segmentContentHashResult->segment;
            // In direct replica calculate hash
            if ($firstCycle && !$dsConfig->reverse && !isset($segmentContentHashResult->partialSegment)) {
                trigger_error("Calculating new segment hash. Query: ".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter, E_USER_NOTICE);
                if (!($hashRset = $conn->query($replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter))) {
                    throw new \Exception("Error executing hash query:".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter." Error: ".$conn->error);
                }
                trigger_error("Segment hash query return ".$hashRset->num_rows." rows", E_USER_NOTICE);
                if ($hashRset->num_rows > 0) {
                    $obj = $hashRset->fetch_object();
                    trigger_error("new segment size ".$obj->SIZE." hash: ".$obj->HASH, E_USER_NOTICE);
                } else {
                    throw new \Exception("Hash query does not return results. ".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter);
                }
                $hashRset->close();
                trigger_error("Confirming replication segment execution and new hash", E_USER_NOTICE);
                $segmentHash->segment->size=$obj->SIZE;
                $segmentHash->hash = $obj->HASH;
            }
            $replicationDeltaResult->segmentHash = $segmentHash;
            if (isset($segmentContentHashResult->partialSegment))
                $replicationDeltaResult->partialSegment = $segmentContentHashResult->partialSegment;
            $this->repService->confirmReplicationDelta ($replicaJob->jobId, $replicationDeltaResult);
            $firstCycle = false;
        } while (!$segmentFullyProcessed);
        if (isset($rset)) {$rset->close();}
        if ($timeout) {return $timeout;}
        if (isset($segmentContentHashResult->partialSegment)) {
            trigger_error("After all partial segments, confirm full segment", E_USER_NOTICE);
            $segmentHash = new SegmentHash();
            $segmentHash->segment = $segmentContentHashResult->segment;
            if (!$dsConfig->reverse) {
                trigger_error("Calculating new segment hash. Query: ". $replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter, E_USER_NOTICE);
                if (!($hashRset = $conn->query($replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter))) {
                    throw new \Exception("Error executing hash query:".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter." Error: ".$conn->error);
                }
                trigger_error("Segment hash query return ".$hashRset->num_rows." rows", E_USER_NOTICE);
                if ($hashRset->num_rows > 0) {
                    $obj = $hashRset->fetch_object();
                    trigger_error("new segment size ".$obj->SIZE." hash: ".$obj->HASH, E_USER_NOTICE);
                } else {
                    throw new \Exception("Hash query does not return results. ".$replicaJob->segmentHashQuery . $segmentHashDescriptor->segmentFilter);
                }
                $hashRset->close();
                trigger_error("Confirming replication segment execution and new hash", E_USER_NOTICE);
                $segmentHash->segment->size=$obj->SIZE;
                $segmentHash->hash = $obj->HASH;
            }
            $replicationDeltaResult->segmentHash = $segmentHash;
            $replicationDeltaResult->partialSegment = NULL;
            $replicationDeltaResult->operationResult = Array();
            $this->repService->confirmReplicationDelta ($replicaJob->jobId, $replicationDeltaResult);
        }
    }
    
    private function calculateSegmentContentHash($rset, $dsConfig, $maxSize=NULL) /*: iterable*/ /* Array of ContentHashItem */
    {
        $res = array();
        while($row = $rset->fetch_assoc()) {
            $first=true;
            $row_string = "";
            $contentHashItem = new ContentHashItem();
            $contentHashItem->row= array();
            foreach ($dsConfig->pk as $pk)
            {
                $contentHashItem->row[] = $row[$pk];
                if (!$first) {
                    $row_string = $row_string.chr(0).$row[$pk];
                }
                else {
                    $row_string=$row[$pk];
                    $first = false;
                }
            }
            foreach ($dsConfig->attributes as $attribute)
            {
                $row_string = $row_string.chr(0).$row[$attribute];
            }
            $contentHashItem->hash=md5($row_string);
            $res[]=$contentHashItem;
            trigger_error("Row: ".json_encode($contentHashItem->row)." hash: ".$contentHashItem->hash, E_USER_NOTICE);
            if (isset($maxSize) && count($res) >= $maxSize) {
                trigger_error("Skipping content fetching due to max size reached (".$maxSize.")");
                break;
            }
        }
        return $res;
    }
    
    private function validateTargetTable($conn, $dsConfig, $targetTableName) /*: boolean, false if not exists*/
    {
        $describeQuery = "DESCRIBE `".$targetTableName."`";
        $targetStructure = array();
        if (!($rset = $conn->query($describeQuery))) {
            if ($conn->errno == 1146) {
                return false;
            } else {
                throw new \Exception("Error executing describe table:".$describeQuery." Error:".$conn->error);
            }
        }
        if ($rset->num_rows > 0) {
            while($obj = $rset->fetch_object()) {
                $targetStructure[]=$obj;
            }
            $this->checkTargetTableStructure($conn, $targetStructure);
        } else {
            throw new \Exception("Describe table:".$describeQuery." return no data");
        }
        return true;
    }
    
    // TODO. Validate target table structure using a DESCRIBE
    private function checkTargetTableStructure($conn, $targetStructure)
    {
        
    }
    
    private function createTarget($conn, $dsReplica)
    {
        $createTable = $this->repService->generateTargetDSScript($dsReplica->dsReplicaId);
        if (!$conn->query($createTable)) {
            throw new \Exception("Error creating target table:".$createTable." Error:".$conn->error);
        }
    }
    
    private function timedout($initTimestamp, $timeout)
    {
        if (isset($timeout) && isset($initTimestamp))
            return (time()-$initTimestamp) >= $timeout;
        return false;
    }
    protected /*IReplicationService*/ $repService;
}

